unit gfx_dib;
{***********************************************************************
Unit gfx_dib.PAS v1.5 0801

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_dib is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.


    This part of gfx_lib is based on the work of Anders Melander, anders@melander.dk

********************************************************************************}

interface
uses gfx_syswrapper, gfx_errors;
////////////////////////////////////////////////////////////////////////////////
//
//			TDIB Classes
//
//  These classes gives read and write access to TXBitmap's pixel data
//  independently of the Delphi version used.
//
////////////////////////////////////////////////////////////////////////////////
type
  TDIB = class(TObject)
  private
    FBitmap		: TXBitmap;
    FPixelFormat	: TxPixelFormat;
    FDIBInfo		: PxBitmapInfo;
    FDIBBits		: pointer;
    FDIBInfoSize	: integer;
    FDIBBitsSize	: longInt;
    FPalette		: HxPalette;
    FHeight		: integer;
    FWidth		: integer;
  protected
    procedure FreeDIB;
    function GetScanline(Row: integer): pointer; virtual; abstract;
    constructor Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat);
  public
    property Scanline[Row: integer]: pointer read GetScanline;
    property Bitmap: TXBitmap read FBitmap;
    property PixelFormat: TxPixelFormat read FPixelFormat;
  end;

  TDIBReader = class(TDIB)
  private
    FDIBInfo		: PxBitmapInfo;
    FDIBBits		: pointer;
    FDIBInfoSize	: xDWORD;
    FDIBBitsSize	: xDWORD;
    FPalette		: HxPalette;
  protected
    function GetScanline(Row: integer): pointer; override;
    procedure CreateDIB;
    procedure NeedDIB;
  public
    constructor Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat);
    destructor Destroy; override;
  end;

  TDIBWriter = class(TDIB)
  private
  protected
    procedure CreateDIB;
    procedure NeedDIB;
    function GetScanline(Row: integer): pointer; override;
  public
    constructor Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat;
      AWidth, AHeight: integer; APalette: HxPalette);
    destructor Destroy; override;
    procedure UpdateBitmap;
    property Width: integer read FWidth;
    property Height: integer read FHeight;
    property Palette: HxPalette read FPalette write FPalette;
  end;

function AlignBit(Bits, BitsPerPixel, Alignment: Cardinal): Cardinal;

implementation
////////////////////////////////////////////////////////////////////////////////
constructor TDIB.Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat);
begin
  inherited Create;
  FBitmap := ABitmap;
  FPixelFormat := APixelFormat;
  FDIBInfo := nil;
  FDIBBits := nil;
  FHeight := FBitmap.Height;
  FWidth := FBitmap.Width;
end;

procedure TDIB.FreeDIB;
begin
  if (FDIBInfo <> nil) then
    FreeMem(FDIBInfo);
  if (FDIBBits <> nil) then
    xGlobalFreePtr(FDIBBits);
  FDIBInfo := nil;
  FDIBBits := nil;
end;

////////////////////////////////////////////////////////////////////////////////
constructor TDIBReader.Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat);
begin
  inherited Create(ABitmap, APixelFormat);
  FBitmap.Pixelformat:=FPixelFormat;
end;

procedure TDIBReader.CreateDIB;
begin
  try
    xGetDIBSizes(FBitmap.Handle, FDIBInfoSize, FDIBBitsSize);
    GetMem(FDIBInfo, FDIBInfoSize);
    FDIBBits := xGlobalAllocPtr(xGMEM_MOVEABLE, FDIBBitsSize);
    if (FDIBBits = nil) then
      raise ExInvalidGraphicOperation.Create('DIB out of memory');
   except
    FreeDIB;
    raise;
   end;
   xGetDIB(FBitmap.Handle, FBitmap.Palette, FDIBINFO^, FDIBBITS^);
   FHeight:=FDIBInfo.bmiHeader.biHeight;
   FWidth:=FDIBInfo.bmiHeader.biWidth;
end;

destructor TDIBReader.Destroy;
begin
  inherited Destroy;
  if (FDIBInfo <> nil) then
    FreeMem(FDIBInfo);
  if (FDIBBits <> nil) then
    xGlobalFreePtr(FDIBBits);
  FDIBInfo := nil;
  FDIBBits := nil;
end;

procedure TDIBReader.NeedDIB;
begin
  if (FDIBBits = nil) then
    CreateDIB;
end;

function TDIBReader.GetScanline(Row: integer): pointer;
begin
  NeedDIB;
    if (FDIBBits = nil) then
    Error('No DIB');
  with FDIBInfo^.bmiHeader do
  begin
    if (Row < 0) or (Row >= FHeight) then
      raise ExInvalidGraphicOperation.Create('Scan line index out of range');
    xGDIFlush;

    if biHeight > 0 then  // bottom-up DIB
      Row := biHeight - Row - 1;
    Result := PChar(Cardinal(FDIBBits) + Cardinal(Row) * AlignBit(biWidth, biBitCount, 32));
  end;

 // Result := FBitmap.ScanLine[Row];
end;

////////////////////////////////////////////////////////////////////////////////
constructor TDIBWriter.Create(ABitmap: TXBitmap; APixelFormat: TxPixelFormat;
  AWidth, AHeight: integer; APalette: HxPalette);
begin
  inherited Create(ABitmap, APixelFormat);

  // DIB writer only supports 8 or 24 bit bitmaps
  if not(APixelFormat in [xpf8bit, xpf24bit]) then
    Error('Invalid Pixelformat');
  if (AWidth = 0) or (AHeight = 0) then
    Error('Bad dimension');

  FHeight := AHeight;
  FWidth := AWidth;
  FPalette := APalette;
  FDIBInfo := nil;
  FDIBBits := nil;
end;

destructor TDIBWriter.Destroy;
begin
  UpdateBitmap;
  FreeDIB;
  inherited Destroy;
end;

function AlignBit(Bits, BitsPerPixel, Alignment: Cardinal): Cardinal;
begin
  Dec(Alignment);
  Result := ((Bits * BitsPerPixel) + Alignment) and not Alignment;
  Result := Result SHR 3;
end;


function TDIBWriter.GetScanline(Row: integer): pointer;
begin
  NeedDIB;
  if (FDIBBits = nil) then
    Error('No DIB');
  with FDIBInfo^.bmiHeader do
  begin
    if (Row < 0) or (Row >= Height) then
      raise ExInvalidGraphicOperation.Create('Scan line index out of range');
    xGDIFlush;

    if biHeight > 0 then  // bottom-up DIB
      Row := biHeight - Row - 1;
    Result := PChar(Cardinal(FDIBBits) + Cardinal(Row) * AlignBit(biWidth, biBitCount, 32));
  end;
//  Result := FBitmap.ScanLine[Row];

end;

procedure TDIBWriter.CreateDIB;
var
  SrcColors		: WORD;
//  ScreenDC		: HDC;

  // From Delphi 3.02 graphics.pas
  // There is a bug in the ByteSwapColors from Delphi 3.0!
  procedure ByteSwapColors(var Colors; Count: Integer);
  var   // convert RGB to BGR and vice-versa.  TRGBQuad <-> TPaletteEntry
    SysInfo: TxSystemInfo;
  begin
    xGetSystemInfo(SysInfo);
    asm
          MOV   EDX, Colors
          MOV   ECX, Count
          DEC   ECX
          JS    @@END
          LEA   EAX, SysInfo
          CMP   [EAX].TxSystemInfo.wProcessorLevel, 3
          JE    @@386
    @@1:  MOV   EAX, [EDX+ECX*4]
          BSWAP EAX
          SHR   EAX,8
          MOV   [EDX+ECX*4],EAX
          DEC   ECX
          JNS   @@1
          JMP   @@END
    @@386:
          PUSH  EBX
    @@2:  XOR   EBX,EBX
          MOV   EAX, [EDX+ECX*4]
          MOV   BH, AL
          MOV   BL, AH
          SHR   EAX,16
          SHL   EBX,8
          MOV   BL, AL
          MOV   [EDX+ECX*4],EBX
          DEC   ECX
          JNS   @@2
          POP   EBX
      @@END:
    end;
  end;
begin
  FreeDIB;

  if (PixelFormat = xpf8bit) then
    // 8 bit: Header and palette
    FDIBInfoSize := SizeOf(TxBitmapInfoHeader) + SizeOf(TxRGBQuad) * (1 shl 8)
  else
    // 24 bit: Header but no palette
    FDIBInfoSize := SizeOf(TxBitmapInfoHeader);

  // Allocate TBitmapInfo structure
  GetMem(FDIBInfo, FDIBInfoSize);
  try
    FDIBInfo^.bmiHeader.biSize := SizeOf(FDIBInfo^.bmiHeader);
    FDIBInfo^.bmiHeader.biWidth := Width;
    FDIBInfo^.bmiHeader.biHeight := Height;
    FDIBInfo^.bmiHeader.biPlanes := 1;
    FDIBInfo^.bmiHeader.biSizeImage := 0;
    FDIBInfo^.bmiHeader.biCompression := xBI_RGB;

    if (PixelFormat = xpf8bit) then
    begin
      FDIBInfo^.bmiHeader.biBitCount := 8;
      // Find number of colors defined by palette
      if (Palette <> 0) and
        (xGetObject(Palette, sizeof(SrcColors), @SrcColors) <> 0) and
        (SrcColors <> 0) then
      begin
        // Copy all colors...
        xGetPaletteEntries(Palette, 0, SrcColors, FDIBInfo^.bmiColors[0]);
        // ...and convert BGR to RGB
        ByteSwapColors(FDIBInfo^.bmiColors[0], SrcColors);
      end else
        SrcColors := 0;

      // Finally zero any unused entried
      if (SrcColors < 256) then
        FillChar(pointer(LongInt(@FDIBInfo^.bmiColors)+SizeOf(TxRGBQuad)*SrcColors)^,
          256 - SrcColors, 0);
      FDIBInfo^.bmiHeader.biClrUsed := 256;
      FDIBInfo^.bmiHeader.biClrImportant := SrcColors;
    end else
    begin
      FDIBInfo^.bmiHeader.biBitCount := 24;
      FDIBInfo^.bmiHeader.biClrUsed := 0;
      FDIBInfo^.bmiHeader.biClrImportant := 0;
    end;
    FDIBBitsSize := AlignBit(Width, FDIBInfo^.bmiHeader.biBitCount, 32) * Cardinal(abs(Height));

    FDIBBits := xGlobalAllocPtr(xGMEM_MOVEABLE, FDIBBitsSize);
    if (FDIBBits = nil) then
      raise ExInvalidGraphicOperation.Create('DIB out of memory');
  except
    FreeDIB;
    raise;
  end;
end;

procedure TDIBWriter.NeedDIB;
begin
  if (FDIBBits = nil) then
    CreateDIB;
end;

// Convert the DIB created by CreateDIB back to a TXBitmap
procedure TDIBWriter.UpdateBitmap;
var
  Stream		: TxMemoryStream;
  FileSize		: longInt;
  BitmapFileHeader	: TxBitmapFileHeader;
begin
  if (FDIBBits = nil) then
    exit;

  begin
    // Write DIB to a stream in the BMP file format
    Stream := TxMemoryStream.Create;
    try
      // Make room in stream for a TBitmapInfo and pixel data
      FileSize := sizeof(TxBitmapFileHeader) + FDIBInfoSize + FDIBBitsSize;
      Stream.SetSize(FileSize);
      // Initialize file header
      FillChar(BitmapFileHeader, sizeof(TxBitmapFileHeader), 0);
      with BitmapFileHeader do
      begin
        bfType := $4D42; // 'BM' = Windows BMP signature
        bfSize := FileSize; // File size (not needed)
        bfOffBits := sizeof(TxBitmapFileHeader) + FDIBInfoSize; // Offset of pixel data
      end;
      // Save file header
      Stream.Write(BitmapFileHeader, sizeof(TxBitmapFileHeader));
      // Save TBitmapInfo structure
      Stream.Write(FDIBInfo^, FDIBInfoSize);
      // Save pixel data
      Stream.Write(FDIBBits^, FDIBBitsSize);

      // Rewind and load bitmap from stream
      Stream.Position := 0;
      FBitmap.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
  end;
end;

end.
